package de.ozgcloud.xta.client.extension;

import de.ozgcloud.xta.client.XtaClient;
import de.ozgcloud.xta.client.config.XtaClientConfig;
import de.ozgcloud.xta.client.core.WrappedXtaService;
import de.ozgcloud.xta.client.core.WrappedXtaServiceFactory;
import de.ozgcloud.xta.client.exception.XtaClientRuntimeException;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.*;
import de.ozgcloud.xta.client.model.*;
import lombok.SneakyThrows;
import lombok.extern.log4j.Log4j2;

import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static java.util.Collections.emptyList;
import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.fail;

@Log4j2
public class XtaServerSetupExtensionTestUtil {

    public static final XtaIdentifier DEV_READER_CLIENT_IDENTIFIER = XtaIdentifier.builder()
            .value("gae:dev-environment@ozg-cloud.de")
            .category("Generischer Antragsempfänger")
            .name("OZG-Cloud Leser Dev")
            .build();
    public static final XtaIdentifier TEST_READER_CLIENT_IDENTIFIER = XtaIdentifier.builder()
            .value("gae:test-environment@ozg-cloud.de")
            .category("Generischer Antragsempfänger")
            .name("OZG-Cloud Leser Test")
            .build();

    public static final XtaIdentifier TEST_AUTHOR_CLIENT_IDENTIFIER = XtaIdentifier.builder()
            .value("gad:010103000000")
            .category("DMS Schleswig-Holstein")
            .name("Generischer Antragsdienst")
            .build();

    // No  certificate allows sending with this identifier yet Does not work yet
    public static final XtaIdentifier DEV_AUTHOR_IDENTIFIER = XtaIdentifier.builder()
            .category("Generischer Antragsdienst")
            .value("gad:010100100000")
            .name("Generischer Antragsdienst Schleswig-Holstein (Integrator)")
            .build();

    public static final XtaIdentifier AUTHOR_CLIENT_IDENTIFIER = TEST_AUTHOR_CLIENT_IDENTIFIER;
    public static final XtaIdentifier AUTHOR_CLIENT_IDENTIFIER2 = XtaIdentifier.builder()
            .value("ehp:010100100000")
            .category("Engagement- und Hobbyportal (FIM Sender)")
            .name("OSI-Onlinedienst Schleswig-Holstein Versammlungsanzeige Test")
            .build();
    public static final XtaIdentifier AUTHOR_CLIENT_IDENTIFIER3 = XtaIdentifier.builder()
            .value("ehp:010200100000")
            .category("Engagement- und Hobbyportal (FIM Sender2)")
            .name("OSI-Onlinedienst Hamburg Versammlungsanzeige Test")
            .build();

    public static final XtaIdentifier READER_CLIENT_IDENTIFIER1 = DEV_READER_CLIENT_IDENTIFIER;
    public static final XtaIdentifier READER_CLIENT_IDENTIFIER2 = TEST_READER_CLIENT_IDENTIFIER;
    public static final XtaIdentifier READER_CLIENT_IDENTIFIER3 = XtaIdentifier.builder()
            .value("vbe:010510440100")
            .category("Versammlungsbehörde (FIM Empfänger)")
            .name("Kreisordnungsbehörde Dithmarschen")
            .build();

    public static XtaClientConfig.XtaClientConfigBuilder createClientConfigBuilder() {
        return XtaClientConfig.builder()
                .clientIdentifiers(emptyList())
                .logSoapRequests(true)
                .logSoapResponses(true);
    }

    @SneakyThrows
    public static String sendTestMessage(XtaClient client, XtaMessage message) {
        try {
            log.info("Sending from author {} to reader {}.", message.metaData().authorIdentifier(), message.metaData().readerIdentifier());
            var transportReport = client.sendMessage(message);
            return transportReport.metaData().messageId();
        } catch (XtaClientRuntimeException e) {
            var cause = e.getCause();
            if (cause instanceof ParameterIsNotValidException) {
                logCodeFehlerNummer(((ParameterIsNotValidException) cause).getFaultInfo().getErrorCode());
            } else if (cause instanceof PermissionDeniedException) {
                logCodeFehlerNummer(((PermissionDeniedException) cause).getFaultInfo().getErrorCode());
            } else if (cause instanceof MessageSchemaViolationException) {
                logCodeFehlerNummer(((MessageSchemaViolationException) cause).getFaultInfo().getErrorCode());
            } else if (cause instanceof XTAWSTechnicalProblemException) {
                logCodeFehlerNummer(((XTAWSTechnicalProblemException) cause).getFaultInfo().getErrorCode());
            }
            throw e;
        }
    }

    private static void logCodeFehlerNummer(CodeFehlernummer fehlernummer) {
        log.error("[{}] {}", fehlernummer.getCode(), fehlernummer.getName());
    }

    @SneakyThrows
    public static void failIfAnyMessagePending(XtaClientConfig config, XtaIdentifier clientId) {
        var wrappedService = createWrappedService(config);
        var result = wrappedService.getStatusList(clientId, 1);
        if (!result.messages().isEmpty()) {
            fail("Expect no pending xta-messages for reader %s! Ensure that the mailbox is empty before running this test.".formatted(
                    clientId.value()));
        }
    }

    @SneakyThrows
    public static void closeAllMessages(XtaClientConfig config, XtaIdentifier clientId) {
        var wrappedService = createWrappedService(config);

        var result = wrappedService.getStatusList(clientId, 100);
        var messageIds = result.messages().stream()
                .map(XtaMessageMetaData::messageId)
                .toList();
        closeMessagesById(config, clientId, messageIds);
    }

    @SneakyThrows
    public static void closeMessagesById(XtaClientConfig config, XtaIdentifier clientId, List<String> messageIds) {
        var wrappedService = createWrappedService(config);
        for (var messageId : messageIds) {
            wrappedService.close(messageId, clientId);
        }
    }

    @SneakyThrows
    private static WrappedXtaService createWrappedService(XtaClientConfig config) {
        return WrappedXtaServiceFactory.from(config).create();
    }

    public static XtaMessageMetaData withoutMessageIdAndSize(XtaMessageMetaData metaData) {
        return metaData.toBuilder()
                .messageId(null)
                .messageSize(null)
                .build();
    }

    public static XtaMessageMetaData withoutMessageSize(XtaMessageMetaData metaData) {
        return metaData.toBuilder()
                .messageSize(null)
                .build();
    }

    public record MessagesAssert(List<XtaMessage> processedMessages) {

        public MessagesAssert containExactlyInAnyOrder(XtaMessage... messages) {
            try {
                containExactlyInAnyOrderRaw(messages);
            } catch (AssertionError | RuntimeException e) {
                log.error("Messages do not exactly contain excepted messages!");
                throw e;
            }
            return this;
        }

        private void containExactlyInAnyOrderRaw(XtaMessage... messages) {
            var messageMetaData = Arrays.stream(messages)
                    .map(XtaMessage::metaData)
                    .toArray(XtaMessageMetaData[]::new);
            containMetaDataExactlyInAnyOrder(messageMetaData);

            assertEqualMessageFileWithoutContent(messages);
            assertEqualAttachmentFilesWithoutContent(messages);

            assertEqualContentOfMessageFile(messages);
            assertEqualContentOfAttachmentFiles(messages);
        }

        private void assertEqualMessageFileWithoutContent(XtaMessage... messages) {
            // ignoring size since it may be null before sending
            assertThat(processedMessages)
                    .extracting(XtaMessage::messageFile)
                    .extracting(this::withoutContentAndSize)
                    .containsExactlyInAnyOrderElementsOf(Arrays.stream(messages)
                            .map(XtaMessage::messageFile)
                            .map(this::withoutContentAndSize)
                            .toList());
        }

        private void assertEqualAttachmentFilesWithoutContent(XtaMessage... messages) {
            // ignoring size since it may be null before sending
            assertThat(processedMessages)
                    .extracting(XtaMessage::attachmentFiles)
                    .extracting(this::filesWithoutContentAndSize)
                    .containsExactlyInAnyOrderElementsOf(Arrays.stream(messages)
                            .map(XtaMessage::attachmentFiles)
                            .map(this::filesWithoutContentAndSize)
                            .toList());
        }

        private List<XtaFile> filesWithoutContentAndSize(List<XtaFile> fileList) {
            return fileList.stream().map(this::withoutContentAndSize).toList();
        }

        private XtaFile withoutContentAndSize(XtaFile xtaFile) {
            return xtaFile.toBuilder()
                    .content(null)
                    .size(null)
                    .build();
        }

        private void assertEqualContentOfMessageFile(XtaMessage... messages) {
            var listOfMessageFileContents = processedMessages.stream()
                    .map(XtaMessage::messageFile)
                    .map(XtaServerSetupExtensionTestUtil::readBytesOfXtaFile)
                    .toList();
            var exceptedListOfMessageFileContents = Arrays.stream(messages)
                    .map(XtaMessage::messageFile)
                    .map(XtaServerSetupExtensionTestUtil::readBytesOfXtaFile)
                    .toList();
            assertThat(listOfMessageFileContents)
                    .extracting(b -> b.length)
                    .containsExactlyInAnyOrderElementsOf(exceptedListOfMessageFileContents.stream().map(b -> b.length).toList());
            assertThat(listOfMessageFileContents).containsExactlyInAnyOrderElementsOf(exceptedListOfMessageFileContents);
        }

        private void assertEqualContentOfAttachmentFiles(XtaMessage... messages) {
            var attachmentFileContents = processedMessages.stream()
                    .map(XtaMessage::attachmentFiles)
                    .map(XtaServerSetupExtensionTestUtil::readBytesOfXtaFiles)
                    .toList();
            var exceptedAttachmentFileContents = Arrays.stream(messages)
                    .map(XtaMessage::attachmentFiles)
                    .map(XtaServerSetupExtensionTestUtil::readBytesOfXtaFiles)
                    .toList();
            assertEqualLengthOfAttachmentFiles(attachmentFileContents, exceptedAttachmentFileContents);
            assertEqualRawContentOfAttachmentFiles(attachmentFileContents, exceptedAttachmentFileContents);
        }

        private void assertEqualLengthOfAttachmentFiles(List<List<byte[]>> attachmentFileContents,
                                                        List<List<byte[]>> exceptedAttachmentFileContents) {
            Function<List<byte[]>, List<Integer>> contentLengths = contentList -> contentList.stream().map(b -> b.length).toList();
            assertThat(attachmentFileContents)
                    .extracting(contentLengths)
                    .containsExactlyInAnyOrderElementsOf(exceptedAttachmentFileContents.stream().map(contentLengths).toList());
        }

        private void assertEqualRawContentOfAttachmentFiles(List<List<byte[]>> attachmentFileContents,
                                                            List<List<byte[]>> exceptedAttachmentFileContents) {
            assertThat(attachmentFileContents)
                    .usingElementComparator(this::compareAttachmentFiles)
                    .containsExactlyInAnyOrderElementsOf(exceptedAttachmentFileContents);
        }

        private int compareAttachmentFiles(List<byte[]> a, List<byte[]> b) {
            var sizeComparison = Integer.compare(a.size(), b.size());
            return sizeComparison != 0
                    ? sizeComparison
                    : IntStream.range(0, a.size())
                    .map(i -> Arrays.compare(a.get(i), b.get(i)))
                    .filter(i -> i != 0)
                    .findFirst()
                    .orElse(0);
        }

        public void containMetaDataExactlyInAnyOrder(XtaMessageMetaData... messageMetaDataItems) {
            try {
                // Assert equal message counts
                assertThat(processedMessages).hasSize(messageMetaDataItems.length);

                // Assert equal metadata (ignoring message id and size since they should be null before sending)
                assertThat(processedMessages)
                        .extracting(XtaMessage::metaData)
                        .extracting(XtaServerSetupExtensionTestUtil::withoutMessageIdAndSize)
                        .containsExactlyInAnyOrderElementsOf(Arrays.stream(messageMetaDataItems)
                                .map(XtaServerSetupExtensionTestUtil::withoutMessageIdAndSize)
                                .toList());
            } catch (AssertionError | RuntimeException e) {
                log.error("Messages do not exactly contain excepted metadata!");
                throw e;
            }
        }
    }

    public static MessagesAssert assertThatMessages(List<XtaMessage> processedMessages) {
        return new MessagesAssert(processedMessages);
    }

    public record TransportReportsAssert(List<XtaTransportReport> transportReports) {
        public TransportReportsAssert reportExactlyFor(List<XtaMessage> processedMessages) {
            try {
                reportExactlyForRaw(processedMessages);
            } catch (AssertionError | RuntimeException e) {
                log.error("TransportReports do not exactly match messages metadata!");
                throw e;
            }
            return this;
        }

        private void reportExactlyForRaw(List<XtaMessage> processedMessages) {
            assertThat(transportReports).hasSize(processedMessages.size());
            assertEqualMessageId(processedMessages);
            assertEqualMessageMetadata(processedMessages);
        }

        private void assertEqualMessageId(List<XtaMessage> processedMessages) {
            assertThat(transportReports)
                    .extracting(XtaTransportReport::metaData)
                    .extracting(XtaMessageMetaData::messageId)
                    .containsExactlyElementsOf(processedMessages.stream()
                            .map(XtaMessage::metaData)
                            .map(XtaMessageMetaData::messageId)
                            .toList());
        }

        private void assertEqualMessageMetadata(List<XtaMessage> processedMessages) {
            // ignoring size since it may be null for transport report (due to message closed?)
            assertThat(transportReports)
                    .extracting(XtaTransportReport::metaData)
                    .extracting(XtaServerSetupExtensionTestUtil::withoutMessageSize)
                    .containsExactlyElementsOf(processedMessages.stream()
                            .map(XtaMessage::metaData)
                            .map(XtaServerSetupExtensionTestUtil::withoutMessageSize)
                            .toList());
        }

        public TransportReportsAssert haveExactlyClosedStatusFor(String... messageIds) {
            try {
                haveExactlyClosedStatusForRaw(messageIds);
            } catch (AssertionError | RuntimeException e) {
                log.error("TransportReports do not have excepted closed status for messageIds!");
                throw e;
            }
            return this;
        }

        private void haveExactlyClosedStatusForRaw(String... messageIds) {
            var setOfMessageIds = Arrays.stream(messageIds).collect(Collectors.toSet());

            assertThat(transportReports)
                    .allMatch(transportReport ->
                            isClosed(transportReport) == expectIsClosed(transportReport, setOfMessageIds)

                    );
        }

        private boolean expectIsClosed(XtaTransportReport transportReport, Set<String> messageIds) {
            return messageIds.contains(transportReport.metaData().messageId());
        }

        private boolean isClosed(XtaTransportReport transportReport) {
            return !transportReport.status().equals(XtaMessageStatus.OPEN);
        }
    }

    public static TransportReportsAssert assertThatTransportReports(List<XtaTransportReport> transportReports) {
        return new TransportReportsAssert(transportReports);
    }

    public static boolean hasLogLineContaining(String logLine) {
        return StaticStringListAppender.getLogLines().stream()
                .anyMatch(line -> line.contains(logLine));
    }

    private static List<byte[]> readBytesOfXtaFiles(List<XtaFile> xtaFiles) {
        return xtaFiles.stream()
                .map(XtaServerSetupExtensionTestUtil::readBytesOfXtaFile)
                .toList();
    }

    @SneakyThrows
    private static byte[] readBytesOfXtaFile(XtaFile xtaFile) {
        try (var inputStream = xtaFile.content().getInputStream()) {
            return inputStream.readAllBytes();
        }
    }

}
