package de.ozgcloud.xta.client.mapper;

import static de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory.*;
import static de.ozgcloud.xta.client.factory.MsgBoxResponseTypeTestFactory.*;
import static de.ozgcloud.xta.client.factory.TransportReportTestFactory.*;
import static de.ozgcloud.xta.client.factory.XtaFileTestFactory.*;
import static java.time.temporal.ChronoUnit.*;
import static org.assertj.core.api.Assertions.*;

import java.math.BigInteger;

import org.apache.commons.io.IOUtils;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.EnumSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mapstruct.factory.Mappers;

import de.ozgcloud.xta.client.factory.ContentTypeTestFactory;
import de.ozgcloud.xta.client.factory.GenericContentContainerTestFactory;
import de.ozgcloud.xta.client.factory.LookupServiceResponseTestFactory;
import de.ozgcloud.xta.client.factory.MessageMetaDataTestFactory;
import de.ozgcloud.xta.client.factory.MsgBoxResponseTypeTestFactory;
import de.ozgcloud.xta.client.factory.MsgStatusListTypeTestFactory;
import de.ozgcloud.xta.client.factory.TransportReportTestFactory;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.ContentType;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.GenericContentContainer;
import de.ozgcloud.xta.client.gen.v3.de.xoev.transport.xta.x211.TransportReport;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgBoxResponseType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2008.x05.transport.MsgStatusListType;
import de.ozgcloud.xta.client.gen.v3.eu.osci.ws.x2014.x10.transport.MessageMetaData;
import de.ozgcloud.xta.client.model.XtaFile;
import de.ozgcloud.xta.client.model.XtaMessage;
import de.ozgcloud.xta.client.model.XtaMessageMetaData;
import de.ozgcloud.xta.client.model.XtaMessageStatus;
import de.ozgcloud.xta.client.model.XtaTransportReport;
import lombok.SneakyThrows;

class ResponseMapperTest {
	private ResponseMapper mapper;

	@BeforeEach
	void setup() {
		mapper = Mappers.getMapper(ResponseMapper.class);
	}

	@DisplayName("map xta message meta listing")
	@Nested
	class TestMapXtaMessageMetaListing {
		private MsgStatusListType msgStatusListType;
		private MsgBoxResponseType msgBoxResponseTypeWithPendingMessages;
		private MsgBoxResponseType msgBoxResponseTypeWithNoPendingMessages;

		@BeforeEach
		void setup() {
			msgStatusListType = MsgStatusListTypeTestFactory.create();
			msgBoxResponseTypeWithPendingMessages = MsgBoxResponseTypeTestFactory
					.createWithPendingMessages();
			msgBoxResponseTypeWithNoPendingMessages = MsgBoxResponseTypeTestFactory
					.createWithNoPendingMessages();
		}

		@DisplayName("should map pending message count")
		@Test
		void shouldMapPendingMessageCount() {
			var result = mapper.mapXtaMessageMetaListing(
					msgStatusListType, msgBoxResponseTypeWithPendingMessages);

			assertThat(result.pendingMessageCount()).isEqualTo(ITEMS_PENDING);
		}

		@DisplayName("should map pending message count to zero if missing")
		@Test
		void shouldMapMoreMessagesToFalseWithNoPendingMessages() {
			var result = mapper.mapXtaMessageMetaListing(
					msgStatusListType, msgBoxResponseTypeWithNoPendingMessages);

			assertThat(result.pendingMessageCount()).isEqualTo(BigInteger.ZERO);
		}

		@DisplayName("should map meta data")
		@Test
		void shouldMapMetaData() {
			var result = mapper.mapXtaMessageMetaListing(
					msgStatusListType, msgBoxResponseTypeWithPendingMessages);

			var messageIds = result.messages().stream()
					.map(XtaMessageMetaData::messageId)
					.toList();
			assertThat(messageIds)
					.containsExactly(MESSAGE_ID, MESSAGE_ID2, MESSAGE_ID3);
		}
	}

	@DisplayName("map xta message meta")
	@Nested
	class TestMapXtaMessageMeta {

		private final MessageMetaData messageMetaData = MessageMetaDataTestFactory.create();

		@DisplayName("should map message id")
		@Test
		void shouldMapMessageId() {
			var result = doMapping();

			assertThat(result.messageId()).isEqualTo(MESSAGE_ID);
		}

		@DisplayName("should map message type code")
		@Test
		void shouldMapMessageTypeCode() {
			var result = doMapping();

			assertThat(result.messageTypeCode()).isEqualTo(MESSAGE_TYPE_CODE);
		}

		@DisplayName("should map service")
		@Test
		void shouldMapService() {
			var result = doMapping();

			assertThat(result.service()).isEqualTo(MESSAGE_SERVICE);
		}

		@DisplayName("should map business scenario code")
		@Test
		void shouldMapBusinessScenarioCode() {
			var result = doMapping();

			assertThat(result.businessScenarioCode()).isEqualTo(BUSINESS_SCENARIO_CODE);
		}

		@DisplayName("should map business scenario code with undefined code")
		@Test
		void shouldMapUndefinedBusinessScenarioCode() {
			var result = mapper.mapXtaMessageMetaData(MessageMetaDataTestFactory.createWithDefined(false));

			assertThat(result.businessScenarioCode()).isEqualTo(UNDEFINED_BUSINESS_SCENARIO_CODE);
		}

		@DisplayName("should map business scenario list uri")
		@Test
		void shouldMapBusinessScenarioListUri() {
			var result = doMapping();

			assertThat(result.businessScenarioListUri()).isEqualTo(BUSINESS_SCENARIO_LIST_URI);
		}

		@DisplayName("should map business scenario list version id")
		@Test
		void shouldMapBusinessScenarioListVersionId() {
			var result = doMapping();

			assertThat(result.businessScenarioListVersionId()).isEqualTo(BUSINESS_SCENARIO_LIST_VERSION_ID);
		}

		@DisplayName("should map message type payload schema")
		@Test
		void shouldMapMessageTypePayloadSchema() {
			var result = doMapping();

			assertThat(result.messageTypePayloadSchema()).isEqualTo(MESSAGE_TYPE_PAYLOAD_SCHEMA);
		}

		@DisplayName("should map author identifier")
		@Test
		void shouldMapAuthorIdentifier() {
			var result = doMapping();

			assertThat(result.authorIdentifier()).isEqualTo(AUTHOR_IDENTIFIER);
		}

		@DisplayName("should map reader identifier")
		@Test
		void shouldMapReaderIdentifier() {
			var result = doMapping();

			assertThat(result.readerIdentifier()).isEqualTo(READER_IDENTIFIER);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = doMapping();

			assertThat(result.messageSize()).isEqualTo(MESSAGE_SIZE);
		}

		private XtaMessageMetaData doMapping() {
			return mapper.mapXtaMessageMetaData(messageMetaData);
		}

	}

	@DisplayName("is service available")
	@Nested
	class TestIsServiceAvailable {
		@DisplayName("should return")
		@ParameterizedTest
		@ValueSource(booleans = { true, false })
		void shouldReturnTrue(boolean serviceAvailable) {
			var result = mapper.isServiceAvailable(LookupServiceResponseTestFactory.create(serviceAvailable));

			assertThat(result).isEqualTo(serviceAvailable);
		}

		@DisplayName("should return false with null")
		@Test
		void shouldReturnFalseWithNull() {
			var result = mapper.isServiceAvailable(LookupServiceResponseTestFactory.create(null));

			assertThat(result).isFalse();
		}
	}

	@DisplayName("map xta file")
	@Nested
	class TestMapXtaFile {

		private final ContentType contentType = ContentTypeTestFactory.create();

		@DisplayName("should map content")
		@Test
		@SneakyThrows
		void shouldMapContent() {
			var result = doMapping();

			assertThat(IOUtils.toByteArray(result.content().getInputStream())).isEqualTo(XTA_FILE_CONTENT);
		}

		@DisplayName("should map content description")
		@Test
		void shouldMapContentDescription() {
			var result = doMapping();

			assertThat(result.contentDescription()).isEqualTo(XTA_FILE_CONTENT_DESCRIPTION);
		}

		@DisplayName("should map content type")
		@Test
		void shouldMapContentType() {
			var result = doMapping();

			assertThat(result.contentType()).isEqualTo(XTA_FILE_CONTENT_TYPE);
		}

		@DisplayName("should map encoding")
		@Test
		void shouldMapEncoding() {
			var result = doMapping();

			assertThat(result.encoding()).isEqualTo(XTA_FILE_ENCODING);
		}

		@DisplayName("should map name")
		@Test
		void shouldMapName() {
			var result = doMapping();

			assertThat(result.name()).isEqualTo(XTA_FILE_NAME);
		}

		@DisplayName("should map id")
		@Test
		void shouldMapId() {
			var result = doMapping();

			assertThat(result.id()).isEqualTo(XTA_FILE_ID);
		}

		@DisplayName("should map language")
		@Test
		void shouldMapLanguage() {
			var result = doMapping();

			assertThat(result.language()).isEqualTo(XTA_FILE_LANGUAGE);
		}

		@DisplayName("should map size")
		@Test
		void shouldMapSize() {
			var result = doMapping();

			assertThat(result.size()).isEqualTo(XTA_FILE_SIZE);
		}

		private XtaFile doMapping() {
			return mapper.mapXtaFile(contentType);
		}
	}

	@DisplayName("map xta message")
	@Nested
	class TestMapXtaMessage {

		private final GenericContentContainer genericContentContainer = GenericContentContainerTestFactory.create();
		private final MessageMetaData metaData = MessageMetaDataTestFactory.create();

		@DisplayName("should map meta data")
		@Test
		void shouldMapMetaData() {
			var result = doMapping();

			assertThat(result.metaData())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapXtaMessageMetaData(metaData));
		}

		@DisplayName("should map message file")
		@Test
		void shouldMapMessageFile() {
			var result = doMapping();

			assertThat(result.messageFile())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapXtaFile(genericContentContainer.getContentContainer().getMessage()));
		}

		@DisplayName("should map attachment files")
		@Test
		void shouldMapAttachmentFiles() {
			var result = doMapping();

			var attachmentFileIds = result.attachmentFiles().stream()
					.map(XtaFile::id)
					.toList();
			assertThat(attachmentFileIds).containsExactly(XTA_FILE_ID_2, XTA_FILE_ID_3);
		}

		private XtaMessage doMapping() {
			return mapper.mapXtaMessage(genericContentContainer, metaData);
		}
	}

	@DisplayName("map xta transport report")
	@Nested
	class TestMapXtaTransportReport {

		private final TransportReport transportReport = TransportReportTestFactory.create();

		@DisplayName("should map meta data")
		@Test
		void shouldMapMetaData() {
			var result = doMapping();

			assertThat(result.metaData())
					.usingRecursiveComparison()
					.isEqualTo(mapper.mapXtaMessageMetaData(transportReport.getMessageMetaData()));
		}

		@DisplayName("should map report time")
		@Test
		void shouldMapReportTime() {
			var result = doMapping();

			assertThat(result.reportTime())
					.isEqualTo(REPORT_TIME.truncatedTo(MILLIS));
		}

		private XtaTransportReport doMapping() {
			return mapper.mapXtaTransportReport(transportReport);
		}

		@DisplayName("should map status")
		@ParameterizedTest
		@EnumSource(XtaMessageStatus.class)
		void shouldMapStatus(XtaMessageStatus status) {
			var result = mapper.mapXtaTransportReport(TransportReportTestFactory.createWithStatus(status));

			assertThat(result.status())
					.isEqualTo(status);
		}

	}
}